#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>

// #include "fwupload_ppdev.h"
#include "fwupload_direct.h"

// #define DEBUG( bla... ) printf( bla );
#define DEBUG( bla... )

int init( int fd, char *fn, u_int32_t size );
int send_block( int fd, unsigned char *data, size_t len );
double gettime();

int main( int argc, char *argv[] )
{
	struct stat fdat;
	int pfd, ifd, n, i, done;
	unsigned char rdbuf[10240];
	double starttime;
	double speed = 0;
	
	printf( "fwupload - Upload images through parallel port to WYSE Winterm devices\n\n" );
	
	if( argc < 2 )
	{
		fprintf( stderr, "Usage: %s <image> [<device>]\n", argv[0] );
		return( 1 );
	}
	
	if( ( ifd = open( argv[1], O_RDONLY ) ) < 0 )
	{
		perror( "open_fw" );
		return( 1 );
	}
	
	if( fstat( ifd, &fdat ) != 0 )
	{
		perror( "stat" );
		return( 1 );
	}
	
	pfd = init_bits( argc >= 3 ? argv[2] : NULL );
	
	printf( "Turn on device to start flashing process... " );
	fflush( stdout );
	if( !init( pfd, argv[1], fdat.st_size ) )
	{
		fprintf( stderr, "Handshake error\n" );
		return( 1 );
	}
	printf( "OK!\n" );
	
	i = done = 0;
	starttime = gettime();
	while( ( n = read( ifd, rdbuf, 10240 ) ) > 0 )
	{
		i = 0;
		while( i < n )
			i += send_block( pfd, rdbuf + i, n - i );
		
		done += i;
		
		if( gettime() != starttime )
			speed = done / ( gettime() - starttime ) / 1024;
		
		printf( "Sent %8d/%8d bytes (%3.1f%%, %2.2f KB/s)\r",
		        done, (int) fdat.st_size, 100.0 * done / fdat.st_size,
		        speed );
		fflush( stdout );
	}
	
	printf( "%d bytes sent. (%2.2f KB/s)                       \n", done, speed );
	
	close( ifd );
	
	return( 0 );
}

char handshake_readsend( int fd, char tosend );
unsigned char readbyte( int fd );
void sendbyte( int fd, unsigned char data );

int init( int fd, char *fn, u_int32_t size )
{
	char *handshake_string = "- Geef poot -", *s;
	char hsbuf[16] = { 0 };
	char fnbuf[16] = { 0 };
	char lastret = 0, rv = 0;
	int i;
	
	for( s = fn; *s; s ++ )
		if( *s == '/' )
			fn = s + 1;
	
	if( strlen( fn ) > 12 )
	{
		fprintf( stderr, "Image filename must be in 8.3 format\n" );
		return( 0 );
	}
	
	for( i = 0; fn[i]; i ++ )
		fnbuf[i] = toupper( fn[i] );
	
	for( i = 0; i < 17; i ++ )
		handshake_readsend( fd, 0xaa );
	
	/* Send the handshake string (the contents doesn't matter that
	   much, it seems) and wait for "OK". */
	s = handshake_string;
	while( 1 )
	{
		if( !*s )
			s = handshake_string;
		
		rv = handshake_readsend( fd, *s );
		
		if( lastret == 'O' && rv == 'K' )
			break;
		
		lastret = rv;
		s ++;
	}
	
	/* Probably for synchronization. */
	for( i = 0; i < 12; i ++ )
		hsbuf[i] = readbyte( fd );
	
	/* I have no idea what this is yet, but the original seems to send
	   zeroes. */
	send_block( fd, (void*) hsbuf + 12, 4 );
	
	/* Originally this comes from a DOS FCB, the last four bytes contain
	   the current file offset, but does the Winterm care? */
	send_block( fd, (void*) fnbuf, 16 );
	
	/* It certainly cares about this. :-) After this the progress
	   indicator will appear. */
	send_block( fd, (void*) &size, 4 );
	/* FIXME: Endianness issue */
	
	return( 1 );
}

int send_block( int fd, unsigned char *data, size_t len )
{
	size_t n;
	int i;
	u_int32_t *arr, cs;
	char *ack = "XMIT";
	static unsigned char id = 1, tmp;
	unsigned char r1, r2;
	
	/* Won't send more than 0x200 bytes per block */
	n = len > 0x200 ? 0x200 : len;
	
	DEBUG( "send_block for %d->%d bytes\n", len, n );
	
	for( i = 0; i < n; i ++ )
		sendbyte( fd, data[i] );
	
	/* Send a checksum. After this, the Winterm will probably figure out
	   if a resend is necessary. */
	cs = 0;
	arr = (void*) data;
	for( i = 0; i < ( n >> 2 ); i ++ )
	{
		/* FIXME: Endianness issue */
		cs -= arr[i];
	}
	
	sendbyte( fd, ( cs       ) & 0xff );
	sendbyte( fd, ( cs >>  8 ) & 0xff );
	sendbyte( fd, ( cs >> 16 ) & 0xff );
	sendbyte( fd, ( cs >> 24 ) & 0xff );
	
	/* Wait for the ACK string "XMIT", which should normally come
	   immediately. */
	i = 0;
	while( 1 )
	{
		if( ( tmp = readbyte( fd ) ) == ack[i] )
		{
			i ++;
			if( ack[i] == 0 )
			{
				r1 = readbyte( fd );
				r2 = readbyte( fd );
				
				DEBUG( "%02x ^ %02x == %02x, id %02x\n", r1, r2, r1 ^ r2, id );
				
				/* If r1 == ~r2, everything is probably okay.
				   Otherwise, wait for another ACK. */
				if( ( r1 ^ r2 ) == 0xff )
					break;
			}
			else
			{
				/* Looks good, but we're not there yet. */
				continue;
			}
		}
		else
		{
			DEBUG( "Unexpected character during ACK: %c\n", tmp );
		}
		
		i = 0;
	}
	
	/* r1 contains the "id" (8-bit value, starts counting at 1) of the
	   last block that was received correctly, so if it doesn't match,
	   we have to send again. */
	if( r1 == id )
	{
		DEBUG( "Block sent successfully.\n" );
		id ++;
		return( n );
	}
	else
	{
		DEBUG( "Going to retry block id=%02x\n", id );
		return( send_block( fd, data, len ) );
	}
}

/* Besides the fact that this one both reads and sends, it also doesn't halt
   when there's no data coming in. */
char handshake_readsend( int fd, char tosend )
{
	unsigned char st1, st2;
	
	sendbits_1( fd, tosend );
	usleep( 1500 );
	st1 = hs_readbits_1( fd );
	
	sendbits_2( fd, tosend >> 4 );
	usleep( 1500 );
	st2 = hs_readbits_2( fd );
	
	return( ( ( st2 << 4 ) & 0xf0 ) | ( st1 & 0x0f ) );
}

unsigned char readbyte( int fd )
{
	char st1, st2;
	
	sendbits_1( fd, 0 );
	st1 = readbits_1( fd );
	
	sendbits_2( fd, 0 );
	st2 = readbits_2( fd );
	
	return( ( st2 << 4 ) | st1 );
}

void sendbyte( int fd, unsigned char data )
{
	sendbits_1( fd, data );
	readbits_1( fd );
	sendbits_2( fd, data >> 4 );
	readbits_2( fd );
}

/* More convenient, and more precision than time(). */
double gettime()
{
	struct timeval time[1];
	
	gettimeofday( time, 0 );
	return( (double) time->tv_sec + (double) time->tv_usec / 1000000 );
}
