#include <linux/ppdev.h>
#include <linux/parport.h>

int init_bits( char *device )
{
	int fd;
	
	if( ( fd = open( device ? device : "/dev/parport0", O_RDWR ) ) < 0 )
	{
		perror( "open_parport" );
		exit( 1 );
	}
	
	if( ioctl( fd, PPCLAIM ) == -1 )
	{
		perror( "PPCLAIM" );
		exit( 1 );
	}
	
	return( fd );
}

inline void sendbits_1( int fd, unsigned char data )
{
	data &= 0xef;
	ioctl( fd, PPWDATA, &data );
	data |= 0x10;
	ioctl( fd, PPWDATA, &data );
}

inline void sendbits_2( int fd, unsigned char data )
{
	data |= 0x10;
	ioctl( fd, PPWDATA, &data );
	data &= 0xef;
	ioctl( fd, PPWDATA, &data );
}

inline unsigned char readbits_1( int fd )
{
	unsigned char st;
	
	while( 1 )
	{
		ioctl( fd, PPRSTATUS, &st );
		if( !( st & PARPORT_STATUS_BUSY ) )
			return( ( st >> 3 ) & 0x0f );
	}
}

inline unsigned char readbits_2( int fd )
{
	unsigned char st;
	
	while( 1 )
	{
		ioctl( fd, PPRSTATUS, &st );
		if( st & PARPORT_STATUS_BUSY )
			return( ( st >> 3 ) & 0x0f );
	}
}

inline unsigned char hs_readbits_1( int fd )
{
	unsigned char st;
	
	ioctl( fd, PPRSTATUS, &st );
	if( st & PARPORT_STATUS_BUSY )
		return( 0 );
	else
		return( st >> 3 );
}

inline unsigned char hs_readbits_2( int fd )
{
	unsigned char st;
	
	ioctl( fd, PPRSTATUS, &st );
	if( !( st & PARPORT_STATUS_BUSY ) )
		return( 0 );
	else
		return( st >> 3 );
}
